#
# Copyright 2014-2025 Thomas Rabaix <thomas.rabaix@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import ioc
from ioc.misc import Dict, deepcopy 
import os
import unittest

current_dir = os.path.dirname(os.path.realpath(__file__))

class HelperTest(unittest.TestCase):
    def test_build(self):
        container = ioc.build([
            "%s/../fixtures/services.yml" % current_dir
        ], parameters={'inline': 'parameter'})

        self.assertEqual(6, len(container.services))
        self.assertEqual(container.get('foo').fake, container.get('fake'))
        self.assertEqual('argument 1', container.get('fake').mandatory)

        self.ok = True
        self.arg2 = True

        fake = container.get('fake')
        self.assertEqual(True, fake.ok)
        self.assertEqual("arg", fake.arg2)

        self.assertTrue(container.get('foo').weak_reference == container.get('weak_reference'))

        self.assertEqual('the argument 1', container.parameters.get('foo.foo'))
        self.assertEqual('parameter', container.parameters.get('inline'))

    def test_deepcopy(self):
        values = [
            {'sad': 1},
            ('%tuple%', 2)
        ]

        for value in values:
            self.assertEqual(value, deepcopy(value))


class DictTest(unittest.TestCase):

    def test_dict(self):
        d = Dict({'key': 'value'})

        self.assertEqual('value', d.get('key'))
        self.assertEqual(None, d.get('key.fake'))
        self.assertEqual('default', d.get('key.fake', 'default'))

        config = Dict()
        managers = config.get_dict('managers', {'foo': 'bar'})

        self.assertEqual(managers.get('foo'), 'bar')

    def test_dict_iterator(self):
        d = Dict({'key': 'value'})

        for key, value in d.iteritems():
            self.assertEqual(key, 'key')
            self.assertEqual(value, 'value')

    def test_all(self):
        d = Dict({'key': 'value'})
        self.assertEqual(d.all(), {'key': 'value'})

        d = Dict({'key': Dict({'value': 'foo'})})
        self.assertEqual(d.all(), {'key': {'value': 'foo'}})

        d = Dict({'key': Dict({'value': ['foo', 'bar']})})
        self.assertEqual(d.all(), {'key': {'value': ['foo', 'bar']}})



#
# Copyright 2014-2025 Thomas Rabaix <thomas.rabaix@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from ioc.component import Definition, Extension

import mailer

class ExtraMailer(mailer.Mailer):
    def create(self, **kwargs):
        return mailer.Message(**kwargs)

class Extension(Extension):
    def load(self, config, container_builder):
        kwargs = {
            'host': config.get('host', "localhost"),
            'port': config.get('port', 0),
            'use_tls': config.get('use_tls', False),
            'usr': config.get('user', None),
            'pwd': config.get('password', None),
            # 'use_ssl': config.get('use_ssl', False),
        }

        container_builder.add('ioc.extra.mailer', Definition('ioc.extra.mailer.di.ExtraMailer', kwargs=kwargs))



#
# Copyright 2014-2025 Thomas Rabaix <thomas.rabaix@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, Callable, Optional, Union
import logging

class Event(object):
    _propagation_stopped = False

    def __init__(self, data: Optional[dict[str, Any]] = None) -> None:
        self.data = data or {}

    def stop_propagation(self) -> None:
        self._propagation_stopped = True

    def is_propagation_stop(self) -> bool:
        return self._propagation_stopped

    def get(self, name: str) -> Any:
        return self.data[name]

    def set(self, name: str, value: Any) -> None:
        self.data[name] = value

    def has(self, name: str) -> bool:
        return name in self.data

class Dispatcher(object):
    def __init__(self, logger: Optional[logging.Logger] = None) -> None:
        self.listeners: dict[str, list[Callable]] = {}
        self.logger = logger

    def dispatch(self, name: str, event: Optional[Union[Event, dict[str, Any]]] = None) -> Event:
        if isinstance(event, dict):
            event = Event(event)

        event = event or Event()

        if self.logger:
            self.logger.debug("event.dispatch: %s" % name)

        if name not in self.listeners:
            return event

        for listener in self.get_listeners(name):
            if self.logger:
                self.logger.debug("event.dispatch: %s to %s" % (name, listener))

            listener(event)

            if event.stop_propagation():
                if self.logger:
                    self.logger.debug("event.dispatch: %s is stopped" % name)

                break

        return event

    def get_listeners(self, name: str) -> list[Callable]:
        """
        Return the callables related to name
        """        
        return list(map(lambda listener: listener[0], self.listeners[name]))

    def add_listener(self, name: str, listener: Callable, priority: int = 0) -> None:
        """
        Add a new listener to the dispatch
        """
        if name not in self.listeners:
            self.listeners[name] = []

        self.listeners[name].append((listener, priority))

        # reorder event
        self.listeners[name].sort(key=lambda listener: listener[1], reverse=True)

    def remove_listener(self, name: str, listener: Callable) -> None:
        if name not in self.listeners:
            return

        self.listeners[name] = [item for item in self.listeners[name] if item != listener]

    def remove_listeners(self, name: str) -> None:
        if name in self.listeners:
            self.listeners[name] = []



#
# Copyright 2014-2025 Thomas Rabaix <thomas.rabaix@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import ioc.loader, ioc.component, ioc.exceptions
import os, datetime

class Extension(ioc.component.Extension):
    def load(self, config, container_builder):

        path = os.path.dirname(os.path.abspath(__file__))

        loader = ioc.loader.YamlLoader()
        loader.load("%s/resources/config/flask.yml" % path, container_builder)

        app = config.get_dict('app', {})

        container_builder.parameters.set('ioc.extra.flask.app.name', app.get('name', ''))
        
        container_builder.parameters.set('ioc.extra.flask.app.static_folder', app.get('static_folder', 'static'))
        container_builder.parameters.set('ioc.extra.flask.app.static_path', app.get('static_path', ''))
        container_builder.parameters.set('ioc.extra.flask.app.static_url_path', app.get('static_url_path', 'static'))
        container_builder.parameters.set('ioc.extra.flask.app.instance_path', app.get('instance_path', 'templates'))
        container_builder.parameters.set('ioc.extra.flask.app.template_folder', app.get('template_folder', ''))
        container_builder.parameters.set('ioc.extra.flask.app.port', app.get('port', 8080))
        container_builder.parameters.set('ioc.extra.flask.app.instance_relative_config', app.get('instance_relative_config', False))

        self.configure_app_config(config, container_builder)
        self.configure_blueprint(config, container_builder)

    def configure_app_config(self, config, container_builder):

        defaults = {
            'DEBUG':                         False,
            'TESTING':                       False,
            'PROPAGATE_EXCEPTIONS':          None,
            'PRESERVE_CONTEXT_ON_EXCEPTION': None,
            'SECRET_KEY':                    None,
            'PERMANENT_SESSION_LIFETIME':    datetime.timedelta(days=31),
            'USE_X_SENDFILE':                False,
            'LOGGER_NAME':                   None,
            'SERVER_NAME':                   None,
            'APPLICATION_ROOT':              None,
            'SESSION_COOKIE_NAME':           'session',
            'SESSION_COOKIE_DOMAIN':         None,
            'SESSION_COOKIE_PATH':           None,
            'SESSION_COOKIE_HTTPONLY':       True,
            'SESSION_COOKIE_SECURE':         False,
            'MAX_CONTENT_LENGTH':            None,
            'SEND_FILE_MAX_AGE_DEFAULT':     12 * 60 * 60,  # 12 hours
            'TRAP_BAD_REQUEST_ERRORS':       False,
            'TRAP_HTTP_EXCEPTIONS':          False,
            'PREFERRED_URL_SCHEME':          'http',
            'JSON_AS_ASCII':                 True,
            'JSON_SORT_KEYS':                True,
            'JSONIFY_PRETTYPRINT_REGULAR':   True,
        }

        c = config.get_dict('config', {})

        for name, value in defaults.items():
            container_builder.parameters.set('ioc.extra.flask.app.%s' % name, c.get(name, value))

        for name, value in c.iteritems():
            container_builder.parameters.set('ioc.extra.flask.app.%s' % name, c.get(name, value))
            defaults[name] = value

        container_builder.parameters.set('ioc.extra.flask.app.config', defaults)

    def configure_blueprint(self, config, container_builder):
        definition = container_builder.get('ioc.extra.flask.app')

        for id, kwargs in config.get_dict('blueprints', {}).iteritems():
            definition.add_call('register_blueprint', [ioc.component.Reference(id)], kwargs.all())

    def post_build(self, container_builder, container):
        """
        This method make sure the flask configuration is fine, and 
        check the if ioc.extra.jinja2 service is available. If so, the 
        flask instance will use this service, by keeping the flask template 
        loader and the one registered at the jinja2
        """
        app = container.get('ioc.extra.flask.app')

        app.config.update(container_builder.parameters.get('ioc.extra.flask.app.config'))

        if container.has('ioc.extra.jinja2'):
            # This must be an instance of jinja.ChoiceLoader
            # This code replace the flask specific jinja configuration to use
            # the one provided by the ioc.extra.jinja2 code
            jinja2 = container.get('ioc.extra.jinja2')

            jinja2.loader.loaders.append(app.create_global_jinja_loader())

            for name, value in app.jinja_env.globals.items():
                if name not in jinja2.globals:
                    jinja2.globals[name] = value                

            for name, value in app.jinja_env.filters.items():
                if name not in jinja2.filters:
                    jinja2.filters[name] = value                

            app.jinja_env = jinja2



#
# Copyright 2014-2025 Thomas Rabaix <thomas.rabaix@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import unittest

from ioc.extra.jinja2.helper import JinjaHelper
from ioc.component import Container
class JinjaHelperTest(unittest.TestCase):
    def test_get_parameter(self):

        container = Container()
        container.parameters.set('hello', 'world')

        helper = JinjaHelper(container)

        self.assertEqual('world', helper.get_parameter('hello'))
        self.assertEqual(None, helper.get_parameter('fake'))
        self.assertEqual('for real', helper.get_parameter('fake', 'for real'))



#
# Copyright 2014-2025 Thomas Rabaix <thomas.rabaix@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import unittest
import ioc.component, ioc.exceptions
import tests.ioc_test.service

class DefinitionTest(unittest.TestCase):
    def test_init(self):
        definition = ioc.component.Definition()
        self.assertIsNone(definition.clazz)
        self.assertEqual(0, len(definition.arguments))

    def test_tag(self):
        definition = ioc.component.Definition()
        definition.add_tag('jinja.filter')

        self.assertFalse(definition.has_tag('salut'))
        self.assertTrue(definition.has_tag('jinja.filter'))

        self.assertEqual([{}], definition.get_tag('jinja.filter'))

class ParameterHolderTest(unittest.TestCase):
    def test_init(self):
        parameter_holder = ioc.component.ParameterHolder()

        self.assertEqual(0, len(parameter_holder.all()))

    def test_item(self):
        parameter_holder = ioc.component.ParameterHolder()
        
        parameter_holder.set('foo', 'bar')

        self.assertEqual(1, len(parameter_holder.all()))

        self.assertEqual('bar', parameter_holder.get('foo'))
        parameter_holder.remove('foo')

        self.assertEqual(0, len(parameter_holder.all()))

    def test_missing_parameter(self):

        parameter_holder = ioc.component.ParameterHolder()

        with self.assertRaises(ioc.exceptions.UnknownParameter):
            parameter_holder.get('error')
        
class ParameterResolverTest(unittest.TestCase):
    def test_init(self):
        ioc.component.ParameterResolver()

    def test_parameters(self):
        holder = ioc.component.ParameterHolder()
        holder['bonjour'] = 'hello'
        holder['le_monde'] = 'world'

        parameter_resolver = ioc.component.ParameterResolver()
        
        self.assertEqual("hello", parameter_resolver.resolve("%bonjour%", holder))
        self.assertEqual("hello world", parameter_resolver.resolve("%bonjour% %le_monde%", holder))
        self.assertEqual(['hello world', 'hello world'], parameter_resolver.resolve(["%bonjour% %le_monde%", "%bonjour% %le_monde%"], holder))        

    def test_parameter_types(self):
        holder = ioc.component.ParameterHolder()
        parameter_resolver = ioc.component.ParameterResolver()
        
        self.assertEqual(1, parameter_resolver.resolve(1, holder))
        self.assertEqual(1.0, parameter_resolver.resolve(1.0, holder))
        self.assertEqual(True, parameter_resolver.resolve(True, holder))
    
    def test_replace_array(self):
        holder = ioc.component.ParameterHolder()
        holder.set('array', [4, 2])

        parameter_resolver = ioc.component.ParameterResolver()

        self.assertEqual([4, 2], parameter_resolver.resolve("%array%", holder))

    def test_replace_tuple(self):
        holder = ioc.component.ParameterHolder()
        holder.set('tuple', "salut")

        parameter_resolver = ioc.component.ParameterResolver()

        self.assertEqual(("salut", 2), parameter_resolver.resolve(("%tuple%", 2), holder))

    def test_escaping(self):
        holder = ioc.component.ParameterHolder()
        holder.set('bonjour', 'hello')
        holder.set('le_monde', 'world')

        parameter_resolver = ioc.component.ParameterResolver()

        self.assertEqual("%hello", parameter_resolver.resolve("%%%bonjour%", holder))
        self.assertEqual("%hello world %", parameter_resolver.resolve("%%%bonjour% %le_monde% %", holder))

        # Recurive parameters ?? => not now
        # holder['foo'] = 'bar'
        # holder['baz'] = '%%%foo% %foo%%% %%foo%% %%%foo%%%'
        # self.assertEqual("%%bar bar%% %%foo%% %%bar%%", parameter_resolver.resolve('%baz%', holder))

    def test_nested_parameters(self):
        holder = ioc.component.ParameterHolder()
        holder.set('bonjour', 'hello')
        holder.set('le_monde', 'world %exclamation%')
        holder.set('exclamation', '!')

        parameter_resolver = ioc.component.ParameterResolver()

        self.assertEqual("hello world !", parameter_resolver.resolve("%bonjour% %le_monde%", holder))

    def test_nested_parameters_recursive(self):
        holder = ioc.component.ParameterHolder()
        holder.set('bonjour', 'hello %le_monde%')
        holder.set('le_monde', '%bonjour% world')

        parameter_resolver = ioc.component.ParameterResolver()

        with self.assertRaises(ioc.exceptions.RecursiveParameterResolutionError):
            parameter_resolver.resolve("%bonjour% %le_monde%", holder)


class ContainerTest(unittest.TestCase):
    def setUp(self):
        self.container = ioc.component.Container()

    def test_add(self):
        self.container.add('myid', {})
        self.container.add('myid', {})
        self.container.add('myid.2', {})

        self.assertEqual(2, len(self.container.services))

    def test_get(self):
        self.container.add('myid', {})
        self.assertEqual({}, self.container.get('myid'))

        with self.assertRaises(ioc.exceptions.UnknownService):
            self.container.get('fake')

class ContainerBuilderTest(unittest.TestCase):
    def setUp(self):
        self.container = ioc.component.ContainerBuilder()

    def test_get_class(self):
        with self.assertRaises(AttributeError):
            self.container.get_class(ioc.component.Definition('tests.ioc_test.test_component.Fake'))

        definition = ioc.component.Definition('tests.ioc_test.service.Fake', [True], {'param': 'salut'})

        c = self.container.get_class(definition)
        
        self.assertEqual(c.__name__, tests.ioc_test.service.Fake.__name__)

    def test_get_instance(self):
        definition = ioc.component.Definition('tests.ioc_test.service.Fake', [True], {'param': 'salut'})
        container = ioc.component.Container()

        i = self.container.get_instance(definition, container)

        self.assertIs(type(i), tests.ioc_test.service.Fake)
        self.assertEqual(True, i.mandatory)
        self.assertEqual('salut', i.param)

    def test_get_container(self):        
        self.container.add('service.id.1', ioc.component.Definition('tests.ioc_test.service.Fake', [True], {'param': 'salut'}))
        self.container.add('service.id.2', ioc.component.Definition('tests.ioc_test.service.Fake', [False], {'param': 'hello'}))
        self.container.add('service.id.3', ioc.component.Definition('tests.ioc_test.service.Foo', [ioc.component.Reference('service.id.2'), None]))

        container = ioc.component.Container()

        self.container.build_container(container)

        self.assertEqual(5, len(container.services))
        self.assertTrue(container.has('service.id.2'))
        self.assertIsInstance(container.get('service.id.2'), tests.ioc_test.service.Fake)
        self.assertIsInstance(container.get('service.id.3'), tests.ioc_test.service.Foo)

        self.assertEqual(container.get('service.id.3').fake, container.get('service.id.2'))

    def test_cyclic_reference(self):
        self.container.add('service.id.1', ioc.component.Definition('tests.ioc_test.service.Foo', [ioc.component.Reference('service.id.1'), None]))

        container = ioc.component.Container()
        
        with self.assertRaises(ioc.exceptions.CyclicReference):
            self.container.build_container(container)

    def test_get_ids_by_tag(self):
        definition = ioc.component.Definition('tests.ioc_test.service.Foo')
        definition.add_tag('jinja.filter')
        self.container.add('service.id.1', definition)

        self.assertEqual([], self.container.get_ids_by_tag('non_existent_tag'))
        self.assertEqual(['service.id.1'], self.container.get_ids_by_tag('jinja.filter'))

    def test_definition_with_inner_definition(self):

        definition = ioc.component.Definition('tests.ioc_test.service.Fake', arguments=[
            ioc.component.Definition('tests.ioc_test.service.Fake', arguments=[
                ioc.component.Definition('tests.ioc_test.service.Fake', arguments=[1])
            ])
        ])

        self.container.add('foo', definition)

        container = ioc.component.Container()
        self.container.build_container(container)

        self.assertIsInstance(container.get('foo'), tests.ioc_test.service.Fake)
        self.assertIsInstance(container.get('foo').mandatory, tests.ioc_test.service.Fake)
        self.assertIsInstance(container.get('foo').mandatory.mandatory, tests.ioc_test.service.Fake)
        
    def test_reference_with_method(self):
        self.container.add('service.id.1', ioc.component.Definition('tests.ioc_test.service.Fake', [ioc.component.Reference('service.id.2', 'set_ok')]))
        self.container.add('service.id.2', ioc.component.Definition('tests.ioc_test.service.Fake', ['foo']))

        container = ioc.component.Container()

        self.container.build_container(container)

        self.assertEqual(container.get('service.id.1').mandatory, container.get('service.id.2').set_ok)

    def test_exception_for_abstract_definition(self):
        definition = ioc.component.Definition('tests.ioc_test.service.Fake', ['foo'], abstract=True)

        container = ioc.component.Container()

        with self.assertRaises(ioc.exceptions.AbstractDefinitionInitialization):
            self.container.get_service("foo", definition, container)

    def test_abstracted_service_not_in_the_container(self):
        definition = ioc.component.Definition('tests.ioc_test.service.Fake', ['foo'], abstract=True)

        self.container.add('service.id.abstract', definition)

        container = ioc.component.Container()

        self.container.build_container(container)

        self.assertEqual(2, len(container.services))

    def test_create_definition_from_abstract_definition(self):
        self.container.add('service.id.abstract', ioc.component.Definition('tests.ioc_test.service.Fake', ['foo'], abstract=True))

        definition = self.container.create_definition('service.id.abstract')

        self.container.add('service.id.1', definition)

        container = self.container.build_container(ioc.component.Container())
        self.assertEqual(3, len(container.services))



#
# Copyright 2014-2025 Thomas Rabaix <thomas.rabaix@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, Union
import yaml

from ioc.component import Definition, Reference, WeakReference, ContainerBuilder
import ioc.helper
import ioc.exceptions
from . import misc

from .misc import OrderedDictYAMLLoader

class Loader(object):
    def fix_config(self, config: dict[str, Any]) -> 'ioc.helper.Dict':
        for key, value in config.items():
            if isinstance(value, dict):
                config[key] = self.fix_config(value)

        return ioc.helper.Dict(config)

class YamlLoader(Loader):
    def support(self, file: str) -> bool:
        return file[-3:] == 'yml'

    def load(self, file: str, container_builder: ContainerBuilder) -> None:

        content = None
        with open(file, 'r', encoding='utf-8') as f:
            content = f.read()
            
        try:
            data = yaml.load(content, OrderedDictYAMLLoader)
        except yaml.scanner.ScannerError as e:
            raise ioc.exceptions.LoadingError("file %s, \nerror: %s" % (file, e))

        for extension, config in data.items():
            if extension in ['parameters', 'services']:
                continue

            if config is None:
                config = {}

            container_builder.add_extension(extension, self.fix_config(config.copy()))

        if 'parameters' in data:
            for key, value in data['parameters'].items():
                container_builder.parameters.set(key, value)

        if 'services' in data:
            for id, meta in data['services'].items():

                if 'arguments' not in meta:
                    meta['arguments'] = []

                if 'class' not in meta:
                    meta['class'] = None

                if 'kwargs' not in meta:
                    meta['kwargs'] = {}

                if 'calls' not in meta:
                    meta['calls'] = []

                if 'tags' not in meta:
                    meta['tags'] = {}

                if 'abstract' not in meta:
                    meta['abstract'] = False

                definition = Definition(
                    clazz=meta['class'], 
                    arguments=self.set_references(meta['arguments']), 
                    kwargs=self.set_references(meta['kwargs']),
                    abstract=meta['abstract']
                )

                for call in meta['calls']:
                    if len(call) == 0:
                        continue

                    if len(call) == 2:
                        call.append({})

                    if len(call) == 1:
                        call.append([])
                        call.append({})

                    definition.method_calls.append(
                        (call[0], self.set_references(call[1]), self.set_references(call[2]))
                    )

                for tag, options in meta['tags'].items():
                    for option in options:
                        definition.add_tag(tag, option)

                container_builder.add(id, definition)

    def set_reference(self, value: Any) -> Any:
        if misc.is_scalar(value) and value[0:1] == '@':
            if '#' in value:
                id, method = value.split("#")
                return Reference(id[1:], method)

            return Reference(value[1:])

        if misc.is_scalar(value) and value[0:2] == '#@':
            return WeakReference(value[2:])

        if misc.is_iterable(value):
            return self.set_references(value)

        return value

    def set_references(self, arguments: Union[list[Any], dict[str, Any]]) -> Union[list[Any], dict[str, Any]]:
        for pos in misc.get_keys(arguments):
            arguments[pos] = self.set_reference(arguments[pos])

        return arguments



#
# Copyright 2014-2025 Thomas Rabaix <thomas.rabaix@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, Optional
import ioc.component
import ioc.loader
import logging

def build(files: list[str], logger: Optional[logging.Logger] = None, parameters: Optional[dict[str, Any]] = None) -> Any:

    if not logger:
        logger = logging.getLogger('ioc')

    if not parameters:
        parameters = {
            'ioc.debug': False
        }

    container_builder = ioc.component.ContainerBuilder(logger=logger)
        
    loaders = [
        ioc.loader.YamlLoader()
    ]

    logger.debug("Loading files")

    for file in files:
        logger.debug("Search loader for file %s" % file)
        for loader in loaders:
            if not loader.support(file):
                continue

            logger.debug("Found loader %s for file %s" % (loader, file))

            loader.load(file, container_builder)

    container = ioc.component.Container()

    for name, value in parameters.items():
        container_builder.parameters.set(name, value)

    container_builder.build_container(container)

    return container



#
# Copyright 2014-2025 Thomas Rabaix <thomas.rabaix@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import ioc.loader, ioc.component
import os

class Extension(ioc.component.Extension):
    def load(self, config, container_builder):

        path = os.path.dirname(os.path.abspath(__file__))

        loader = ioc.loader.YamlLoader()
        loader.load("%s/resources/config/command.yml" % path, container_builder)

    def post_build(self, container_builder, container):
        command_manager = container.get('ioc.extra.command.manager')

        for id in container_builder.get_ids_by_tag('command'):
            definition = container_builder.get(id)
            for option in definition.get_tag('command'):
                if 'name' not in option:
                    break

                command_manager.add_command(option['name'], container.get(id))



#
# Copyright 2014-2025 Thomas Rabaix <thomas.rabaix@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import unittest
import ioc.locator
import os

current_dir = os.path.dirname(os.path.realpath(__file__))

class FilesystemLocatorTest(unittest.TestCase):

    def test_locate_with_fake_path(self):
        locator = ioc.locator.FileSystemLocator('fake')

        with self.assertRaises(ioc.locator.ResourceNotFound):
            locator.locate('missing file')

    def test_locate(self):
        locator = ioc.locator.FileSystemLocator(current_dir + "/../fixtures")

        self.assertEqual(current_dir + "/../fixtures/services.yml", locator.locate('services.yml'))

class FunctionLocatorTest(unittest.TestCase):
    def test_locate_with_fake_path(self):
        def function(resource):
            return None

        locator = ioc.locator.FunctionLocator(function)

        with self.assertRaises(ioc.locator.ResourceNotFound):
            locator.locate('missing file')

    def test_locate(self):
        def function(resource):
            return "/mypath/%s" % resource

        locator = ioc.locator.FunctionLocator(function)

        self.assertEqual("/mypath/services.yml", locator.locate('services.yml'))

class PrefixLocatorTest(unittest.TestCase):
    def test_locate_with_fake_path(self):
        locator = ioc.locator.PrefixLocator({})

        with self.assertRaises(ioc.locator.ResourceNotFound):
            locator.locate('missing file')

    def test_locate(self):
        locator = ioc.locator.PrefixLocator({
            "app" : ioc.locator.FileSystemLocator(current_dir + "/../fixtures")
        }, ":")

        self.assertEqual(current_dir + "/../fixtures/services.yml", locator.locate('app:services.yml'))

class ChoiceLocatorTest(unittest.TestCase):
    def test_locate(self):
        locator = ioc.locator.ChoiceLocator([
            ioc.locator.FileSystemLocator("/tmp"),
            ioc.locator.FileSystemLocator(current_dir + "/../fixtures"),
        ])

        self.assertEqual(current_dir + "/../fixtures/services.yml", locator.locate('services.yml'))

class PackageLocatorTest(unittest.TestCase):
    def test_locate(self):
        locator = ioc.locator.PackageLocator('tests', 'fixtures')
        self.assertEqual(os.path.realpath(current_dir + "/../fixtures/services.yml"), locator.locate('services.yml'))



#
# Copyright 2014-2025 Thomas Rabaix <thomas.rabaix@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import ioc.loader, ioc.component
import unittest

current_dir = os.path.dirname(os.path.realpath(__file__))

class YamlLoaderTest(unittest.TestCase):

    def test_support(self):
        loader = ioc.loader.YamlLoader()

        self.assertTrue(loader.support('foo.yml'))
        self.assertFalse(loader.support('foo.xml'))

    def test_load(self):
        builder = ioc.component.ContainerBuilder()

        loader = ioc.loader.YamlLoader()
        loader.load("%s/../fixtures/services.yml" % current_dir, builder)

        self.assertEqual(5, len(builder.services))
        self.assertTrue('foo' in builder.services)
        self.assertTrue('fake' in builder.services)

        self.assertIsInstance(builder.get('foo'), ioc.component.Definition)
        self.assertIsInstance(builder.get('fake'), ioc.component.Definition)
        self.assertIsInstance(builder.get('foo').arguments[0], ioc.component.Reference)

        self.assertEqual(2, len(builder.get('fake').method_calls))

        self.assertEqual('set_ok', builder.get('fake').method_calls[0][0])
        self.assertEqual([False], builder.get('fake').method_calls[0][1])
        self.assertEqual({}, builder.get('fake').method_calls[0][2])

        self.assertEqual('set_ok', builder.get('fake').method_calls[1][0])
        self.assertEqual([True], builder.get('fake').method_calls[1][1])
        self.assertEqual({'arg2': 'arg'}, builder.get('fake').method_calls[1][2])

        # test tags
        self.assertEqual(['foo'], builder.get_ids_by_tag('jinja.filter'))

    def test_reference(self):
        loader = ioc.loader.YamlLoader()
        
        arguments = ['@fake', ['@hello', '#@weak_reference'], 1]
        
        loader.set_references(arguments)

        self.assertIsInstance(arguments[0], ioc.component.Reference)
        self.assertIsInstance(arguments[1][0], ioc.component.Reference)
        self.assertIsInstance(arguments[1][1], ioc.component.WeakReference)
        self.assertEqual(arguments[2], 1)

        arguments = {'fake': '@hello', 'boo': ['@fake']}

        arguments = loader.set_references(arguments)

        self.assertIsInstance(arguments['fake'], ioc.component.Reference)
        self.assertIsInstance(arguments['boo'][0], ioc.component.Reference)

        self.assertEqual(arguments['fake'].id, 'hello')

    def test_reference_method(self):
        builder = ioc.component.ContainerBuilder()

        loader = ioc.loader.YamlLoader()
        loader.load("%s/../fixtures/services.yml" % current_dir, builder)

        definition = builder.get('method_reference')

        self.assertIsInstance(definition, ioc.component.Definition)
        self.assertIsInstance(definition.arguments[0], ioc.component.Reference)
        self.assertEqual("fake", definition.arguments[0].id)
        self.assertEqual("set_ok", definition.arguments[0].method)
    
    def test_abstract_service(self):
        builder = ioc.component.ContainerBuilder()

        loader = ioc.loader.YamlLoader()
        loader.load("%s/../fixtures/services.yml" % current_dir, builder)

        self.assertTrue(builder.get('abstract_service').abstract)
        self.assertFalse(builder.get('fake').abstract)



#
# Copyright 2014-2025 Thomas Rabaix <thomas.rabaix@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import unittest
import ioc.event

class EventTest(unittest.TestCase):
    def test_init(self):
        event = ioc.event.Event({'foo': 'bar'})
        self.assertEqual('bar', event.get('foo'))

        with self.assertRaises(KeyError):
            self.assertEqual('bar', event.get('foo2'))

        self.assertFalse(event.has('foo2'))

        event.set('foo2', 'bar')
        self.assertTrue(event.has('foo2'))
        self.assertEqual('bar', event.get('foo2'))

    def test_stop_propagation(self):
        event = ioc.event.Event()

        self.assertFalse(event.is_propagation_stop())
        event.stop_propagation()
        self.assertTrue(event.is_propagation_stop())

def mylistener(event):
    event.set('enter', True)

class EventDispatcherTest(unittest.TestCase):
    def test_init(self):
        ioc.event.Dispatcher()

    def test_listener(self):
        dispatcher = ioc.event.Dispatcher()
        dispatcher.add_listener('node.load', mylistener)

        event = dispatcher.dispatch('node.load', {
            'node': {}
        })

        self.assertIsInstance(event, ioc.event.Event)
        self.assertTrue(event.has('node'))
        self.assertTrue(event.has('enter'))
        self.assertTrue(event.get('enter'))

    def test_remove_listener(self):
        dispatcher = ioc.event.Dispatcher()
        dispatcher.add_listener('node.load', mylistener)
        dispatcher.remove_listeners('node.load')

        event = dispatcher.dispatch('node.load', {
            'node': {}
        })

        self.assertFalse(event.has('enter'))

    def test_get_listener(self):
        listeners = [
            ('event32', 32),
            ('event0', 0),
            ('event1', 1),
            ('event-1', -1)
        ]

        dispatcher = ioc.event.Dispatcher()

        for listener, priority in listeners:
            dispatcher.add_listener('node.load', listener, priority)

        expected = ['event32', 'event1', 'event0', 'event-1']

        self.assertEqual(expected, dispatcher.get_listeners('node.load'))

        



#
# Copyright 2014-2025 Thomas Rabaix <thomas.rabaix@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import ioc.loader, ioc.component
import os
import importlib
import jinja2

class Extension(ioc.component.Extension):
    def load(self, config, container_builder):

        path = os.path.dirname(os.path.abspath(__file__))

        loader = ioc.loader.YamlLoader()
        loader.load("%s/resources/config/jinja.yml" % path, container_builder)

        # register template
        # retrieve all extensions and build a valid MapPrefixLoader dict
        mapping = {}
        for name, config in container_builder.extensions.items():
            mapping[name] = jinja2.FileSystemLoader([
                "%s/resources/%s/templates" % (container_builder.parameters.get('project.root_folder'), name),
                "%s/resources/templates" % os.path.dirname(importlib.import_module(name).__file__),
            ])

        container_builder.parameters.set("ioc.extra.jinja2.loader_mapping", mapping)

    def post_build(self, container_builder, container):
        """
        Register filter and global in jinja environment instance

        IoC tags are:
            - jinja2.filter to register filter, the tag must contain
            a name and a method options
            - jinja2.global to add new global, here globals are functions. 
            The tag must contain a name and a method options
        """
        jinja = container.get('ioc.extra.jinja2')

        for id in container_builder.get_ids_by_tag('jinja2.filter'):
            definition = container_builder.get(id)
            for option in definition.get_tag('jinja2.filter'):
                if 'name' not in option:
                    break

                if 'method' not in option:
                    break

                jinja.filters[option['name']] = getattr(container.get(id), option['method'])

        for id in container_builder.get_ids_by_tag('jinja2.global'):
            definition = container_builder.get(id)
            for option in definition.get_tag('jinja2.global'):

                if 'name' not in option:
                    break

                if 'method' not in option:
                    break                

                jinja.globals[option['name']] = getattr(container.get(id), option['method'])



#
# Copyright 2014-2025 Thomas Rabaix <thomas.rabaix@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import ioc.proxy, ioc.component
import unittest

class FakeService(object):
    p = 42

    def __init__(self):
        self.arg = None

    def method(self):
        return "method"

class ProxyTest(unittest.TestCase):

    def test_support(self):

        fake = FakeService()

        container = ioc.component.Container()
        container.add('fake', fake)

        proxy = ioc.proxy.Proxy(container, 'fake')

        self.assertEqual("method", proxy.method())

        fake.arg = 1

        self.assertEqual(1, proxy.arg)
        self.assertEqual(42, proxy.p)

        self.assertIsInstance(proxy, ioc.proxy.Proxy)
        self.assertIsInstance(proxy, FakeService)



#
# Copyright 2014-2025 Thomas Rabaix <thomas.rabaix@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse

class Command(object):

    def initialize(self, parser):
        pass

    def execute(self, args, output):
        pass

class CommandManager(object):
    def __init__(self, commands=None):
        self.commands = commands or {}

    def add_command(self, name, command):
        parser = argparse.ArgumentParser(prog=name) 

        # default share arguments
        parser.add_argument('--verbose', '-v', action='count', help="verbose level")
        parser.add_argument('--debug', '-d', help="debug mode", action='store_true')
        parser.add_argument('--env', '-e', help="Define the environement")

        command.initialize(parser)

        self.commands[name] = (parser, command)

    def execute(self, argv, stdout):
        argv.pop(0)  # remove the script name

        if len(argv) == 0:  # no argument
            name = 'help'
            argv = []
        else:
            name = argv.pop(0)

        if name not in self.commands:
            name = 'help'

        parser, command = self.commands[name]

        arguments = parser.parse_args(argv)

        r = command.execute(arguments, stdout)

        if r is None:
            r = 0

        return r

class HelpCommand(Command):

    def __init__(self, command_manager):
        self.command_manager = command_manager

    def initialize(self, parser):
        parser.description = 'Display available commands'
        pass

    def execute(self, args, output):
        output.write("Commands available: \n")
        for name, (parser, command) in self.command_manager.commands.items():
            output.write(" > % -20s : %s \n" % (name, parser.description))

        output.write("\n--\nPython IoC - Thomas Rabaix <thomas.rabaix@gmail.com>\n")
